@@ -119,9 +119,9 @@ def _check_njobs(njobs):
 def _check_beta_prior(beta_prior, nchoices, for_ucb=False):
     if beta_prior == 'auto':
         if not for_ucb:
-            out = ( (3.0 / float(nchoices), 4.0), 2 )
+            out = ( (2.0 / np.log2(nchoices), 4.0), 2 )
         else:
-            out = ( (5.0 / float(nchoices), 4.0), 2 )
+            out = ( (3.0 / np.log2(nchoices), 4.0), 2 )
     elif beta_prior is None:
         out = ((1.0,1.0), 0)
     else:
@@ -139,7 +139,7 @@ def _check_smoothing(smoothing):
         return None
     assert len(smoothing) >= 2
     assert (smoothing[0] >= 0) & (smoothing[1] >= 0)
-    assert smoothing[1] > smoothing[0]
+    assert smoothing[1] >= smoothing[0]
     return float(smoothing[0]), float(smoothing[1])
 
 
@@ -151,13 +151,13 @@ def _check_fit_input(X, a, r, choice_names = None):
     assert X.shape[0] == r.shape[0]
     if choice_names is not None:
         a = pd.Categorical(a, choice_names).codes
-        if pd.isnull(a).sum() > 0:
+        if (a < 0).any():
             raise ValueError("Input contains actions/arms that this object does not have.")
     return X, a, r
 
 def _check_X_input(X):
     if (X.__class__.__name__ == 'DataFrame') or isinstance(X, pd.core.frame.DataFrame):
-        X = X.values
+        X = X.to_numpy()
     if isinstance(X, np.matrixlib.defmatrix.matrix):
         warnings.warn("'defmatrix' will be cast to array.")
         X = np.array(X)
@@ -330,9 +330,11 @@ def _gen_random_grad_norms(X, n_pos, n_neg, random_state):
 def _gen_zero_norms(X, n_pos, n_neg):
     return np.zeros((X.shape[0], 2))
 
-def _apply_smoothing(preds, smoothing, counts):
+def _apply_smoothing(preds, smoothing, counts, add_noise, random_state):
     if (smoothing is not None) and (counts is not None):
         preds[:, :] = (preds * counts + smoothing[0]) / (counts + smoothing[1])
+        if add_noise:
+            preds[:, :] += random_state.uniform(low=0., high=1e-12, size=preds.shape)
     return None
 
 def _apply_sigmoid(x):
@@ -343,24 +345,24 @@ def _apply_sigmoid(x):
     return None
 
 def _apply_inverse_sigmoid(x):
-    x[x == 0] = 1e-8
-    x[x == 1] = 1 - 1e-8
+    x[x <= 0] = 1e-8
+    x[x >= 1] = 1. - 1e-8
     if (len(x.shape) == 2):
         x[:, :] = np.log(x / (1.0 - x))
     else:
         x[:] = np.log(x / (1.0 - x))
     return None
 
-def is_from_module(base):
+def is_from_this_module(base):
     return (isinstance(base, _BootstrappedClassifierBase) or
             isinstance(base, _LinUCB_n_TS_single) or
             isinstance(base, _LogisticUCB_n_TS_single) or
-            isinstance(base, _BayesianLogisticRegression) or
             isinstance(base, _TreeUCB_n_TS_single))
 
 def _apply_softmax(x):
     x[:, :] = np.exp(x - x.max(axis=1).reshape((-1, 1)))
     x[:, :] = x / x.sum(axis=1).reshape((-1, 1))
+    x[x > 1] = 1.
     return None
 
 class _FixedPredictor:
@@ -433,13 +435,16 @@ class _RandomPredictor(_FixedPredictor):
 
 class _BootstrappedClassifierBase:
     def __init__(self, base, nsamples, percentile = 80, partialfit = False,
-                 partial_method = "gamma", random_state = 1, njobs = 1):
+                 partial_method = "gamma", ts_byrow = False, ts_weighted = False,
+                 random_state = 1, njobs = 1):
         self.bs_algos = [deepcopy(base) for n in range(nsamples)]
         self.partialfit = partialfit
         self.partial_method = partial_method
         self.nsamples = nsamples
         self.percentile = percentile
         self.njobs = njobs
+        self.ts_byrow = bool(ts_byrow)
+        self.ts_weighted = bool(ts_weighted)
         self.random_state = _check_random_state(random_state)
 
     def fit(self, X, y):
@@ -495,28 +500,39 @@ class _BootstrappedClassifierBase:
         else:
             raise ValueError(_unexpected_err_msg)
 
-    def _score_max(self, X):
+    def _pred_by_sample(self, X):
         pred = np.empty((X.shape[0], self.nsamples))
         Parallel(n_jobs=self.njobs, verbose=0, require="sharedmem")\
                 (delayed(self._assign_score)(sample, pred, X) \
                     for sample in range(self.nsamples))
+        return pred
+
+    def _score_max(self, X):
+        pred = self._pred_by_sample(X)
         return np.percentile(pred, self.percentile, axis=1)
 
     def _score_avg(self, X):
         ### Note: don't try to make it more memory efficient by summing to a single array,
         ### as otherwise it won't be multithreaded.
-        pred = np.empty((X.shape[0], self.nsamples))
-        Parallel(n_jobs=self.njobs, verbose=0, require="sharedmem")\
-                (delayed(self._assign_score)(sample, pred, X) \
-                    for sample in range(self.nsamples))
+        pred = self._pred_by_sample(X)
         return pred.mean(axis = 1)
 
     def _assign_score(self, sample, pred, X):
         pred[:, sample] = self._get_score(sample, X)
 
     def _score_rnd(self, X):
-        chosen_sample = self.random_state.integers(self.nsamples)
-        return self._get_score(chosen_sample, X)
+        if not self.ts_byrow:
+            chosen_sample = self.random_state.integers(self.nsamples)
+            return self._get_score(chosen_sample, X)
+        else:
+            pred = self._pred_by_sample(X)
+            if not self.ts_weighted:
+                return pred[np.arange(X.shape[0]),
+                            self.random_state.integers(self.nsamples, size=X.shape[0])]
+            else:
+                w = self.random_state.random(size = (X.shape[0], self.nsamples))
+                w[:] /= w.sum(axis=0, keepdims=True)
+                return np.einsum("ij,ij->i", w, pred)
 
     def exploit(self, X):
         return self._score_avg(X)
@@ -646,7 +662,7 @@ class _OneVsRest:
                  X, a, r, n,
                  thr, alpha, beta,
                  random_state,
-                 smooth=False, assume_un=False,
+                 smooth=False, noise_to_smooth=True, assume_un=False,
                  partialfit=False, refit_buffer=0, deep_copy=False,
                  force_fit=False, force_counters=False,
                  prev_ovr=None, warm=False,
@@ -654,6 +670,7 @@ class _OneVsRest:
                  njobs=1):
         self.n = n
         self.smooth = smooth
+        self.noise_to_smooth = bool(noise_to_smooth)
         self.assume_un = assume_un
         self.njobs = njobs
         self.force_fit = bool(force_fit)
@@ -710,11 +727,11 @@ class _OneVsRest:
                 for choice in range(self.n):
                     if isinstance(self.algos[choice], _FixedPredictor):
                         self.algos[choice] = deepcopy(base)
-                        if is_from_module(base):
+                        if is_from_this_module(base):
                             self.algos[choice].random_state = self.rng_arm[choice]
             else: 
                 self.algos = [deepcopy(base) for choice in range(self.n)]
-                if is_from_module(base):
+                if is_from_this_module(base):
                     for choice in range(self.n):
                         self.algos[choice].random_state = self.rng_arm[choice]
 
@@ -835,9 +852,9 @@ class _OneVsRest:
         if self.assume_un:
             this_choice = (a == choice)
             arms_w_rew = (r > 0.)
-            yclass = r[this_choice | arms_w_rew]
-            yclass[arms_w_rew & (~this_choice) ] = 0
+            yclass = np.where(arms_w_rew & (~this_choice), np.zeros_like(r), r)
             this_choice = this_choice | arms_w_rew
+            yclass = yclass[this_choice]
         else:
             this_choice = (a == choice)
             yclass = r[this_choice]
@@ -852,7 +869,8 @@ class _OneVsRest:
         Parallel(n_jobs=self.njobs, verbose=0, require="sharedmem")\
                 (delayed(self._decision_function_single)(choice, X, preds, 1) \
                     for choice in range(self.n))
-        _apply_smoothing(preds, self.smooth, self.counters)
+        _apply_smoothing(preds, self.smooth, self.counters,
+                         self.noise_to_smooth, self.random_state)
         return preds
 
     def _decision_function_single(self, choice, X, preds, depth=2):
@@ -892,7 +910,8 @@ class _OneVsRest:
         Parallel(n_jobs=self.njobs, verbose=0, require="sharedmem")\
                 (delayed(self._decision_function_single)(choice, X, preds, 1) \
                     for choice in range(self.n))
-        _apply_smoothing(preds, self.smooth, self.counters)
+        _apply_smoothing(preds, self.smooth, self.counters,
+                         self.noise_to_smooth, self.random_state)
         _apply_inverse_sigmoid(preds)
         _apply_softmax(preds)
         return preds
@@ -902,7 +921,8 @@ class _OneVsRest:
         Parallel(n_jobs=self.njobs, verbose=0, require="sharedmem")\
                 (delayed(self._decision_function_single)(choice, X, preds, 0) \
                     for choice in range(self.n))
-        _apply_smoothing(preds, self.smooth, self.counters)
+        _apply_smoothing(preds, self.smooth, self.counters,
+                         self.noise_to_smooth, self.random_state)
         return preds
 
     def predict(self, X):
@@ -926,7 +946,7 @@ class _OneVsRest:
     def get_n_neg(self, choice):
         return self.beta_counters[2, choice]
 
-    def get_counts(self):
+    def get_nobs_by_arm(self):
         return self.beta_counters[1] + self.beta_counters[2]
 
     def exploit(self, X):
@@ -940,78 +960,10 @@ class _OneVsRest:
     def _exploit_single(self, choice, pred, X):
         pred[:, choice] = self.algos[choice].exploit(X)
 
-
-class _BayesianLogisticRegression:
-    def __init__(self, method='advi', niter=2000, nsamples=50,
-                 mode='ucb', perc=None, random_state=1):
-        #TODO: reimplement with something faster than using PyMC3's black-box methods
-        self.nsamples = nsamples
-        self.niter = niter
-        self.mode = mode
-        self.perc = perc
-        self.method = method
-        self.random_state = _check_random_state(random_state)
-
-    def fit(self, X, y):
-        try:
-            import pymc3 as pm
-        except:
-            msg  = "This method requires PyCM3. "
-            msg += "Can be installed with e.g. 'pip install pymc3'."
-            raise ValueError(msg)
-        import pandas as pd
-        import logging
-        logger = logging.getLogger("pymc3")
-        logger.propagate = False
-        with pm.Model():
-            pm.glm.linear.GLM(X, y, family = 'binomial')
-            if self.method == 'advi':
-                trace = pm.fit(progressbar = False, n = self.niter,
-                               random_seed = self.random_state.integers(
-                                                np.iinfo(np.int32).max + 1
-                                            ))
-            else:
-                trace = pm.sample(progressbar = False, draws = self.niter, tune = 500,
-                                  step = None if self.method == "nuts" else pm.Metropolis(),
-                                  random_seed = self.random_state.integers(
-                                                np.iinfo(np.int32).max + 1
-                                            ))
-        if self.method == 'advi':
-            self.coefs = [i for i in trace.sample(self.nsamples)]
-            self.coefs = pd.DataFrame.from_dict(self.coefs)
-        else:
-            samples_chosen = self.random_state.choice(len(trace),
-                                                      size = self.nsamples,
-                                                      replace = False)
-            self.coefs = [i for i in trace]
-            self.coefs = pd.DataFrame.from_dict(self.coefs)
-            self.coefs = self.coefs.iloc[samples_chosen]
-        self.coefs = self.coefs[ ['Intercept'] + ['x' + str(i) for i in range(X.shape[1])] ]
-        self.intercept = self.coefs['Intercept'].values.reshape((1, -1)).copy()
-        del self.coefs['Intercept']
-        self.coefs = self.coefs.to_numpy().T
-
-    ### TODO: implement 'partial_fit' with stochastic variational inference
-
-    def _predict_all(self, X):
-        pred_all = X.dot(self.coefs) + self.intercept
-        _apply_sigmoid(pred_all)
-        return pred_all
-
-    def predict(self, X, exploit = False):
-        pred = self._predict_all(X)
-        if exploit:
-            return pred.mean(axis = 1)
-        elif self.mode == 'ucb':
-            pred = np.percentile(pred, self.perc, axis=1)
-        elif self.mode == 'ts':
-            pred = pred[:, self.random_state.integers(pred.shape[1])]
-        else:
-            raise ValueError(_unexpected_err_msg)
-        return pred
-
-    def exploit(self, X):
-        return self.predict(X, exploit = True)
+    def reset_attribute(self, attr_name, attr_value):
+        for model in self.algos:
+            if is_from_this_module(model):
+                setattr(model, attr_name, attr_value)
 
 class _LinUCB_n_TS_single:
     def __init__(self, alpha=1.0, lambda_=1.0, fit_intercept=True,
@@ -1056,17 +1008,18 @@ class _LinUCB_n_TS_single:
 
     def exploit(self, X):
         if not self.is_fitted:
-            return 0.
+            return np.zeros(X.shape[0])
         return self.predict(X, exploit = True)
 
 class _LogisticUCB_n_TS_single:
     def __init__(self, lambda_=1., fit_intercept=True, alpha=0.95,
-                 m=1.0, ts=False, sample_unique=False, random_state=1):
-        self.conf_coef = norm_dist.ppf(alpha)
+                 m=1.0, ts=False, ts_from_ci=True, sample_unique=False, random_state=1):
+        self.conf_coef = alpha
         self.m = m
         self.fit_intercept = fit_intercept
         self.lambda_ = lambda_
         self.ts = ts
+        self.ts_from_ci = ts_from_ci
         self.warm_start = True
         self.sample_unique = bool(sample_unique)
         self.random_state = _check_random_state(random_state)
@@ -1077,6 +1030,11 @@ class _LogisticUCB_n_TS_single:
                                         warm_start=True)
         self.Sigma = np.empty((0,0), dtype=np.float64)
 
+    def __setattr__(self, name, value):
+        if (name == "conf_coef"):
+            value = norm_dist.ppf(value / 100.)
+        super().__setattr__(name, value)
+
     def fit(self, X, y, *args, **kwargs):
         if X.shape[0] == 0:
             return self
@@ -1111,8 +1069,27 @@ class _LogisticUCB_n_TS_single:
             Xcsr = None
         return X, Xcsr
 
-    def decision_function(self, X, exploit=False):
-        ### Thompson sampling
+    def predict(self, X, exploit=False):
+        ## TODO: refactor this, merge code from ucb and ts_from_ci
+        if (exploit) and (not self.is_fitted):
+            return np.zeros(X.shape[0])
+
+        ### Thompson sampling, from CI
+        if (self.ts) and (self.ts_from_ci) and (not exploit):
+            if not self.is_fitted:
+                if not issparse(X):
+                    ci = np.sqrt(np.einsum("ij,ij->i", X, X) / self.lambda_)
+                else:
+                    ci = np.sqrt(X.multiply(X).sum(axis=1) / self.lambda_)
+                return self.random_state.normal(size=X.shape[0]) * ci
+            pred = self.model.decision_function(X)
+            X, Xcsr = self._process_X(X)
+            se_sq = _wrapper_double.x_A_x_batch(X, self.Sigma, Xcsr, self.fit_intercept, 1)
+            pred[:] += self.random_state.normal(size=X.shape[0]) * np.sqrt(se_sq.reshape(-1))
+            _apply_sigmoid(pred)
+            return pred
+
+        ### Thompson sampling, from coefficients
         if (self.ts) and (not exploit):
             if self.fit_intercept:
                 coef = np.r_[self.model.coef_.reshape(-1), self.model.intercept_]
@@ -1146,12 +1123,18 @@ class _LogisticUCB_n_TS_single:
                 pred = X.dot(coef[:X.shape[1]])
                 if self.fit_intercept:
                     pred[:] += coef[-1]
+            _apply_sigmoid(pred)
             return pred
 
         ### UCB
         if not self.is_fitted:
-            pred = self.conf_coef * np.sqrt(np.einsum("ij,ij->i", X, X) / self.lambda_)
-            pred[:] += self.random_state.random(size=pred.shape[0]) / 1e5
+            if not issparse(X):
+                se_sq = np.einsum("ij,ij->i", X, X)
+            else:
+                se_sq = X.multiply(X).sum(axis=1)
+            pred = self.conf_coef * np.sqrt(se_sq / self.lambda_)
+            pred[:] += self.random_state.uniform(low=0., high=1e-12, size=pred.shape[0])
+            _apply_sigmoid(pred)
             return pred
 
         pred = self.model.decision_function(X)
@@ -1159,11 +1142,12 @@ class _LogisticUCB_n_TS_single:
             X, Xcsr = self._process_X(X)
             se_sq = _wrapper_double.x_A_x_batch(X, self.Sigma, Xcsr, self.fit_intercept, 1)
             pred[:] += self.conf_coef * np.sqrt(se_sq.reshape(-1))
+        _apply_sigmoid(pred)
         return pred
 
     def exploit(self, X):
         if not self.is_fitted:
-            return 0.
+            return np.zeros(X.shape[0])
         pred = self.model.decision_function(X)
         _apply_sigmoid(pred)
         return pred
@@ -1173,12 +1157,17 @@ class _TreeUCB_n_TS_single:
                  *args, **kwargs):
         self.beta_prior = beta_prior
         self.random_state = random_state
-        self.conf_coef = norm_dist.ppf(alpha)
+        self.conf_coef = alpha
         self.ts = bool(ts)
         self.model = DecisionTreeClassifier(*args, **kwargs)
         self.is_fitted = False
         self.aux_beta = (beta_prior[0], beta_prior[1])
 
+    def __setattr__(self, name, value):
+        if (name == "conf_coef"):
+            value = norm_dist.ppf(value / 100.)
+        super().__setattr__(name, value)
+
     def update_aux(self, y):
         self.aux_beta[0] += (y >  0.).sum()
         self.aux_beta[1] += (y <= 0.).sum()
@@ -1228,7 +1217,7 @@ class _TreeUCB_n_TS_single:
                 return self.random_state.beta(self.aux_beta[0], self.aux_beta[1], size=n)
             else:
                 mean = self.aux_beta[0] / (self.aux_beta[0] + self.aux_beta[1])
-                noise = self.random_state.random(size=n) / 1e6
+                noise = self.random_state.uniform(low=0., high=1e-12, size=n)
                 return mean + noise
 
         pred_node = self.model.apply(X)
